module Tokyo

  def progress
    @progress ||= TTY::ProgressBar.new ':current of :total [:bar]' do |cfg|
      cfg.total = units.map {|u| u.tests.size}.reduce(:+) || 0
      cfg.width = TTY::Screen.width
      cfg.complete = '.'
    end
  end

  def run pattern_or_files = DEFAULT_PATTERN
    specs = 0
    tests = 0
    assertions = 0
    find_files(pattern_or_files).shuffle.each do |file|
      r, w = IO.pipe
      pid = Kernel.fork do
        r.close
        load_file(file)
        progress.log ''
        progress.log cyan(relative_location(file))
        units.shuffle.each do |unit|
          # exceptions raised inside unit#__run__ will be treated as failures and pretty printed.
          # any other exceptions will be treated as implementation errors and ugly printed.
          unit.tests.keys.shuffle.each do |test|
            status = unit.run(test)
            if status.is_a?(Skip)
              w.puts({skip: true, reason: status.reason, caller: status.caller}.to_json)
            else
              unless status == :__tokyo_passed__
                render_failure(unit, unit.tests[test], status)
                Kernel.exit(1)
              end
            end
            progress.advance
          end
        end
        w.puts(totals.to_json)
      end
      _, status = Process.waitpid2(pid)
      Kernel.exit(1) unless status.success?
      w.close
      while line = r.gets
        line = JSON.parse(line)
        if line['skip']
          skips << line
        elsif line['totals']
          specs += line['specs']
          tests += line['tests']
          assertions += line['assertions']
        else
          raise('Incomprehensible message received: %s' % line)
        end
      end
    end
    render_skips
    render_totals(specs, tests, assertions)
  end

  def totals
    {
      totals: true,
      specs: units.select {|u| u.__ancestors__.empty?}.size,
      tests: units.map {|u| u.tests.size}.reduce(:+),
      assertions: total_assertions.size
    }
  end

  def render_totals specs, tests, assertions
    puts
    puts
    puts bold.cyan('           Specs: %i' % specs)
    puts bold.cyan('           Tests: %i' % tests)
    puts bold.cyan('      Assertions: %i' % assertions)
    puts
  end

  def render_failure unit, test_uuid, failure
    indent = ''
    [*unit.__ancestors__, unit].each do |u|
      progress.log indent + u.__identity__
      indent << INDENT
    end
    progress.log indent + test_uuid
    indent << INDENT
    case failure
    when Exception
      render_exception(indent, failure)
    when GenericFailure, AssertionFailure
      render_caller(indent, failure.caller)
      __send__('render_%s' % failure.class.name.split('::').last, indent, failure)
    else
      progress.log(indent + failure.inspect)
    end
    progress.log ''
  end

  def render_exception indent, failure
    progress.log indent + underline.bright_red([failure.class, failure.message]*': ')
    pretty_backtrace(failure).each {|l| progress.log(indent + l)}
  end

  def render_GenericFailure indent, failure
    Array(failure.reason).each {|l| progress.log(indent + l.to_s)}
  end

  def render_AssertionFailure indent, failure
    progress.log indent + cyan('a: ') + pp(failure.object)
    progress.log indent + cyan('b: ') + failure.arguments.map {|a| pp(a)}.join(', ')
  end

  def render_caller indent, caller
    return unless caller
    progress.log indent + underline.bright_red(readline(caller))
  end

  def render_skips
    return if skips.empty?
    puts
    puts bold.magenta('Skips:')
    skips.each do |skip|
      puts '  %s (%s)' % [blue(skip['reason'] || 'skip'), relative_location(skip['caller'])]
    end
    puts
  end
end
